// CSigScan
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2014 Dominik Friedrichs

#include "CSigScan.h"


CSigScan::CSigScan(const char *Signature,
                   const char *Module)
{
  SigAddress = NULL;

  SigData = NULL;
  SigMask = NULL;

  SigLength = 0;

  MemoryBase = NULL;
  MemoryLength = 0;

  SetSignature(Signature);
  SetModule(Module);

  return;
}


CSigScan::~CSigScan(void)
{
  FreeSignatureMemory();
}


bool CSigScan::SetSignature(const char *str_hexsig)
{
  if(!str_hexsig)
    return(false);

  SigAddress = NULL;
  FreeSignatureMemory();

  int siglen = HexSigToBinary(str_hexsig,
                              SigData, SigMask);

  if(siglen <= 0)
  // error or zero length
  {
    FreeSignatureMemory();
    return(false);
  }

  SigLength = siglen;
  return(true);
}


bool CSigScan::SetModule(const char *str_module)
{
  if(!str_module)
    return(false);

  HMODULE module = GetModuleHandle(str_module);

  MemoryBase = NULL;
  MemoryLength = 0;

  SigAddress = NULL;
  MEMORY_BASIC_INFORMATION mem;

  if(!module
  || !VirtualQuery(module, &mem, sizeof(mem)))
    return(false);

  IMAGE_DOS_HEADER *dos = (IMAGE_DOS_HEADER*)
                          mem.AllocationBase;
  IMAGE_NT_HEADERS *pe = (IMAGE_NT_HEADERS*)
                         ((unsigned long)dos+(unsigned long)dos->e_lfanew);

  if(pe->Signature != IMAGE_NT_SIGNATURE)
    return(false);

  MemoryBase = (const unsigned char *)mem.AllocationBase;
  MemoryLength = pe->OptionalHeader.SizeOfImage;

  return(true);
}


bool CSigScan::FindSignature(void)
{
  if(!MemoryBase
  || !SigData || !SigMask)
    return(false);

  SigAddress = NULL;
  unsigned char *match = NULL;

  for(const unsigned char *mem_ptr = MemoryBase,
                          *mem_end = MemoryBase
                                   + MemoryLength
                                   - SigLength;
      mem_ptr <= mem_end; mem_ptr++)
  {
    unsigned int byte;

    for(byte = 0; byte < SigLength; byte++)
    if(SigMask[byte]
    && mem_ptr[byte] != SigData[byte])
      break;

    if(byte != SigLength)
    // incomplete or no match
      continue;

    if(match)
    // disallow multiple matches
      return(false);

    match = (unsigned char *)mem_ptr;
  }

  if(!match)
  // no match at all
    return(false);

  SigAddress = match;
  return(true);
}


void *CSigScan::GetSigAddress(int Offset)
{
  return(SigAddress + Offset);
}


void CSigScan::FreeSignatureMemory(void)
{
  delete [] SigData;
  delete [] SigMask;

  SigData = NULL;
  SigMask = NULL;

  SigLength = 0;
  return;
}


int CSigScan::HexSigToBinary(const char *str,
                             unsigned char *&data,
                             unsigned char *&mask)
{
  if(!str
  || !&data || !&mask)
    return(-1);

  unsigned int len = (strlen(str)+1)/3;

  if(!len)
    return(-1);

  unsigned int byte;

  data = new unsigned char[len];
  mask = new unsigned char[len];

  for(byte = 0; *str; byte++, str += 2)
  {
    if(byte > 0)
    {
      if(byte >= len
      || *str != '\x20')
      // overflow or not whitespace
      {
        byte = 0;
        break;
      }

      // skip spaces inbetween
      str++;
    }

    char hexp[3];
    strncpy(hexp, str, 2);
    hexp[2] = '\0';

    if(!strcmp(hexp, "??"))
    // wildcard byte
    {
      data[byte] = 0x00;
      mask[byte] = 0x00;
      continue;
    }

    if(!IsHexChar(hexp[0])
    || !IsHexChar(hexp[1]))
    // invalid byte
      break;

    data[byte] = (unsigned char)strtoul(hexp, NULL, 16);
    mask[byte] = 0xFF;
  }

  if(byte != len)
  // loop terminated abnormally
    return(-1);

  return(byte);
}


bool CSigScan::IsHexChar(char c)
{
  if((c >= '0' && c <= '9')
  || (c >= 'a' && c <= 'f')
  || (c >= 'A' && c <= 'F'))
    return(true);

  return(false);
}
